@@ -19,15 +19,20 @@
 #ifdef HAVE_ASM_IOCTLS_H
 #include <asm/ioctls.h>
 #endif
+#ifdef HAVE_LINUX_HAYESESP_H
+#include <linux/hayesesp.h>
+#endif
 #include <linux/serial.h>
 
-#define VERSION_STR "2.14"
+#define VERSION_STR "2.15"
 
 char *progname;
 
 int	verbosity = 1;		/* 1 = normal, 0=boot-time, 2=everything */
+				/* -1 == arguments to setserial */
 int	verbose_flag = 0;	/* print results after setting a port */
 int	quiet_flag = 0;
+int	zero_flag = 0;
 
 struct serial_type_struct {
 	int id;
@@ -66,6 +71,12 @@ struct serial_type_struct {
 #define CMD_CONFIG	10
 #define CMD_GETMULTI	11
 #define CMD_SETMULTI	12
+#define CMD_RX_TRIG     13
+#define CMD_TX_TRIG     14
+#define CMD_FLOW_OFF    15
+#define CMD_FLOW_ON     16
+#define CMD_RX_TMOUT    17
+#define CMD_DMA_CHAN    18
 
 #define FLAG_CAN_INVERT	0x0001
 #define FLAG_NEED_ARG	0x0002
@@ -95,8 +106,10 @@ struct flag_type_table {
 	CMD_FLAG,	"pgrp_lockout", ASYNC_PGRP_LOCKOUT, ASYNC_PGRP_LOCKOUT, 2, FLAG_CAN_INVERT,
 #ifdef ASYNC_CALLOUT_NOHUP
 	CMD_FLAG,	"callout_nohup", ASYNC_CALLOUT_NOHUP, ASYNC_CALLOUT_NOHUP, 2, FLAG_CAN_INVERT,
-#endif	
-	
+#endif
+#ifdef ASYNC_LOW_LATENCY
+	CMD_FLAG,	"low_latency", ASYNC_LOW_LATENCY, ASYNC_LOW_LATENCY, 0, FLAG_CAN_INVERT,
+#endif /* ASYNC_LOW_LATENCY */
 	CMD_PORT,	"port",		0,		0,		0, FLAG_NEED_ARG,
 	CMD_IRQ,	"irq",		0,		0,		0, FLAG_NEED_ARG,
 	CMD_DIVISOR,	"divisor",	0,		0,		0, FLAG_NEED_ARG,
@@ -109,8 +122,17 @@ struct flag_type_table {
 	CMD_WAIT2,	"closing_wait2",0,		0,		0, FLAG_NEED_ARG,
 #endif
 	CMD_CONFIG,	"autoconfig",	0,		0,		0, 0,
+	CMD_CONFIG,	"autoconfigure",0,		0,		0, 0,
 	CMD_GETMULTI,	"get_multiport",0,		0,		0, 0,
 	CMD_SETMULTI,	"set_multiport",0,		0,		0, 0,
+#ifdef HAVE_LINUX_HAYESESP_H
+	CMD_RX_TRIG,    "rx_trigger",   0,              0,              0, FLAG_NEED_ARG,
+	CMD_TX_TRIG,    "tx_trigger",   0,              0,              0, FLAG_NEED_ARG,
+	CMD_FLOW_OFF,   "flow_off",     0,              0,              0, FLAG_NEED_ARG,
+	CMD_FLOW_ON,    "flow_on",      0,              0,              0, FLAG_NEED_ARG,
+	CMD_RX_TMOUT,   "rx_timeout",   0,              0,              0, FLAG_NEED_ARG,
+	CMD_DMA_CHAN,   "dma_channel",  0,              0,              0, FLAG_NEED_ARG,
+#endif
 	0,		0,		0,		0,		0, 0,
 };
 	
@@ -162,6 +184,11 @@ void print_flags(struct serial_struct *serinfo,
 	for (p = flag_type_tbl; p->name; p++) {
 		if (p->cmd != CMD_FLAG)
 			continue;
+		if (verbosity == -1) {
+			if ((flags & p->mask) == p->bits)
+				printf(" %s", p->name);
+			continue;
+		}
 		if (verbosity < p->level)
 			continue;
 		if ((flags & p->mask) == p->bits) {
@@ -179,6 +206,36 @@ void print_flags(struct serial_struct *serinfo,
 }
 
 #ifdef TIOCSERGETMULTI
+void print_multiport(char *device, int fd)
+{
+	struct serial_multiport_struct multi;
+
+	if (ioctl(fd, TIOCSERGETMULTI, &multi) < 0)
+		return;
+
+	if (!multi.port1 && !multi.port2 &&
+	    !multi.port3 && !multi.port4 && !multi.port_monitor)
+		return;
+	
+	printf("%s", device);
+	if (multi.port_monitor)
+		printf(" port_monitor 0x%x", multi.port_monitor);
+	if (multi.port1)
+		printf(" port1 0x%x mask1 0x%x match1 0x%x", multi.port1,
+		       multi.mask1, multi.match1);
+	if (multi.port2)
+		printf(" port2 0x%x mask2 0x%x match2 0x%x", multi.port2,
+		       multi.mask2, multi.match2);
+	if (multi.port3)
+		printf(" port3 0x%x mask3 0x%x match3 0x%x", multi.port3,
+		       multi.mask3, multi.match3);
+	if (multi.port4)
+		printf(" port4 0x%x mask4 0x%x match4 0x%x", multi.port4,
+		       multi.mask4, multi.match4);
+	printf("\n");
+}
+
+
 void multiport_usage()
 {
 
@@ -260,27 +317,27 @@ void set_multiport(char *device, int fd, char ***in_arg)
 			continue;
 		}
 		if (strcasecmp(word, "port3") == 0) {
-			multi.port2 = atonum(argument);
+			multi.port3 = atonum(argument);
 			continue;
 		}
 		if (strcasecmp(word, "mask3") == 0) {
-			multi.mask2 = atonum(argument);
+			multi.mask3 = atonum(argument);
 			continue;
 		}
 		if (strcasecmp(word, "match3") == 0) {
-			multi.match2 = atonum(argument);
+			multi.match3 = atonum(argument);
 			continue;
 		}
 		if (strcasecmp(word, "port4") == 0) {
-			multi.port2 = atonum(argument);
+			multi.port4 = atonum(argument);
 			continue;
 		}
 		if (strcasecmp(word, "mask4") == 0) {
-			multi.mask2 = atonum(argument);
+			multi.mask4 = atonum(argument);
 			continue;
 		}
 		if (strcasecmp(word, "match4") == 0) {
-			multi.match2 = atonum(argument);
+			multi.match4 = atonum(argument);
 			continue;
 		}
 		fprintf(stderr, "Unknown keyword %s.\n", word);
@@ -306,6 +363,62 @@ void set_multiport(char *device, int fd, char ***in_arg)
 }
 #endif
 
+#ifdef HAVE_LINUX_HAYESESP_H
+void print_hayesesp(int fd)
+{
+	struct hayes_esp_config esp;
+
+	if (ioctl(fd, TIOCGHAYESESP, &esp) < 0)
+		return;
+
+	printf("\tHayes ESP enhanced mode configuration:\n");
+	printf("\t\tRX trigger level: %d, TX trigger level: %d\n",
+	       (int)esp.rx_trigger, (int)esp.tx_trigger);
+	printf("\t\tFlow off level: %d, Flow on level: %d\n",
+	       (int)esp.flow_off, (int)esp.flow_on);
+	printf("\t\tRX timeout: %u, DMA channel: %d\n\n",
+	       (unsigned int)esp.rx_timeout, (int)esp.dma_channel);
+}
+
+void set_hayesesp(int fd, int cmd, int arg)
+{
+	struct hayes_esp_config esp;
+
+	if (ioctl(fd, TIOCGHAYESESP, &esp) < 0) {
+		printf("\nError: rx_trigger, tx_trigger, flow_off, "
+		       "flow_on, rx_timeout, and dma_channel\n"
+		       "are only valid for Hayes ESP ports.\n\n");
+		exit(1);
+	}
+
+	switch (cmd) {
+	case CMD_RX_TRIG:
+		esp.rx_trigger = arg;
+		break;
+	case CMD_TX_TRIG:
+		esp.tx_trigger = arg;
+		break;
+	case CMD_FLOW_OFF:
+		esp.flow_off = arg;
+		break;
+	case CMD_FLOW_ON:
+		esp.flow_on = arg;
+		break;
+	case CMD_RX_TMOUT:
+		esp.rx_timeout = arg;
+		break;
+	case CMD_DMA_CHAN:
+		esp.dma_channel = arg;
+		break;
+	}
+
+	if (ioctl(fd, TIOCSHAYESESP, &esp) < 0) {
+		printf("Cannot set Hayes ESP info\n");
+		exit(1);
+	}
+}
+#endif
+
 void get_serial(char *device)
 {
 	struct serial_struct serinfo;
@@ -316,6 +429,7 @@ void get_serial(char *device)
 		perror(device);
 		return;
 	}
+	serinfo.reserved_char[0] = 0;
 	if (ioctl(fd, TIOCGSERIAL, &serinfo) < 0) {
 		perror("Cannot get serial info");
 		close(fd);
@@ -323,7 +437,13 @@ void get_serial(char *device)
 	}
 	if (serinfo.irq == 9)
 		serinfo.irq = 2;	/* People understand 2 better than 9 */
-	if (verbosity==2) {
+	if (verbosity==-1) {
+		printf("%s uart %s port 0x%.4x irq %d", device,
+		       serial_type(serinfo.type), serinfo.port,
+		       serinfo.irq);
+		print_flags(&serinfo, ", Flags: ", "");
+		printf("\n");
+	} else if (verbosity==2) {
 		printf("%s, Line %d, UART: %s, Port: 0x%.4x, IRQ: %d\n",
 		       device, serinfo.line, serial_type(serinfo.type),
 		       serinfo.port, serinfo.irq);
@@ -347,6 +467,10 @@ void get_serial(char *device)
 #endif
 		print_flags(&serinfo, "\tFlags: ", "");
 		printf("\n\n");
+
+#ifdef HAVE_LINUX_HAYESESP_H
+		print_hayesesp(fd);
+#endif
 	} else if (verbosity==0) {
 		if (serinfo.type) {
 			printf("%s at 0x%.4x (irq = %d) is a %s",
@@ -385,6 +509,8 @@ void set_serial(char *device, char ** arg)
 		exit(1);
 	}
 	new_serinfo = old_serinfo;
+	if (zero_flag)
+		new_serinfo.flags = 0;
 	while (*arg) {
 		do_invert = 0;
 		word = *arg++;
@@ -484,7 +610,16 @@ void set_serial(char *device, char ** arg)
 			}
 			set_multiport(device, fd, &arg);
 			break;
-			
+#ifdef HAVE_LINUX_HAYESESP_H
+		case CMD_RX_TRIG:
+		case CMD_TX_TRIG:
+		case CMD_FLOW_OFF:
+		case CMD_FLOW_ON:
+		case CMD_RX_TMOUT:
+		case CMD_DMA_CHAN:
+			set_hayesesp(fd, p->cmd, atonum(*arg++));
+			break;
+#endif
 		default:
 			fprintf(stderr, "Internal error: unhandled cmd #%d\n", p->cmd);
 			exit(1);
@@ -538,7 +673,9 @@ void usage()
 {
 	fprintf(stderr, "setserial Version %s\n\n", VERSION_STR);
 	fprintf(stderr,
-		"usage: %s serial-device [cmd1 [arg]] ... \n\n", progname);
+		"usage:\t %s serial-device -abqvVWz [cmd1 [arg]] ... \n", 
+		progname);
+	fprintf(stderr, "\t %s -g [-abGv] device1 ...\n\n", progname);
 	fprintf(stderr, "Available commands: (* = Takes an argument)\n");
 	fprintf(stderr, "\t\t(^ = can be preceded by a '^' to turn off the option)\n");
 fprintf(stderr, "\t* port\t\tset the I/O port\n");
@@ -561,7 +698,7 @@ fprintf(stderr, "\t* port\t\tset the I/O port\n");
 	fprintf(stderr, "\t\t\t\tafter the reciver is disabled\n");
 #endif
 	fprintf(stderr, "\t^ fourport\tconfigure the port as an AST Fourport\n");
-	fprintf(stderr, "\t  autoconfigure\tautomatically configure the serial port\n");
+	fprintf(stderr, "\t  autoconfig\tautomatically configure the serial port\n");
 	fprintf(stderr, "\t^ auto_irq\ttry to determine irq during autoconfiguration\n");
 	fprintf(stderr, "\t^ skip_test\tskip UART test during autoconfiguration\n");
 	fprintf(stderr, "\n");
@@ -576,10 +713,23 @@ fprintf(stderr, "\t* port\t\tset the I/O port\n");
 	fprintf(stderr, "\t^ split_termios Use separate termios for callout and dailin lines\n");
 #endif	
 	fprintf(stderr, "\t^ hup_notify\tNotify a process blocked on opening a dial in line\n");
-	fprintf(stderr, "\t  get_multiport\tDisplay the multiport configuration\n");
-	fprintf(stderr, "\t  set_multiport\tSet the multiport configuration\n");
 	fprintf(stderr, "\t\t\t\twhen a process has finished using a callout\n");
 	fprintf(stderr, "\t\t\t\tline by returning EAGAIN to the open.\n");
+#ifdef ASYNC_LOW_LATENCY
+	fprintf(stderr, "\t^ low_latency\tMinimize receive latency at the cost of greater\n");
+	fprintf(stderr, "\t\t\t\tCPU utilization.\n");
+#endif
+	fprintf(stderr, "\t  get_multiport\tDisplay the multiport configuration\n");
+	fprintf(stderr, "\t  set_multiport\tSet the multiport configuration\n");
+	fprintf(stderr, "\n");
+#ifdef HAVE_LINUX_HAYESESP_H
+	fprintf(stderr, "\t* rx_trigger\tSet RX trigger level\n");
+	fprintf(stderr, "\t* tx_trigger\tSet TX trigger level\n");
+	fprintf(stderr, "\t* flow_off\tSet hardware flow off level\n");
+	fprintf(stderr, "\t* flow_on\tSet hardware flow on level\n");
+	fprintf(stderr, "\t* rx_timeout\tSet receive timeout\n");
+	fprintf(stderr, "\t* dma_channel\tSet DMA channel");
+#endif
 	fprintf(stderr, "\n");
 	fprintf(stderr, "\t  spd_hi\tuse 56kb instead of 38.4kb\n");
 	fprintf(stderr, "\t  spd_vhi\tuse 115kb instead of 38.4kb\n");
@@ -602,7 +752,7 @@ main(int argc, char **argv)
 	progname = argv[0];
 	if (argc == 1)
 		usage();
-	while ((c = getopt(argc, argv, "abgqvVW")) != EOF) {
+	while ((c = getopt(argc, argv, "abgGqvVWz")) != EOF) {
 		switch (c) {
 		case 'a':
 			verbosity = 2;
@@ -619,12 +769,18 @@ main(int argc, char **argv)
 		case 'g':
 			get_flag++;
 			break;
+		case 'G':
+			verbosity = -1;
+			break;
 		case 'V':
 			fprintf(stderr, "setserial version %s\n", VERSION_STR);
 			exit(0);
 		case 'W':
 			wild_intr_flag++;
 			break;
+		case 'z':
+			zero_flag++;
+			break;
 		default:
 			usage();
 		}
