@@ -5,5 +5,5 @@
  * redistributed under the GNU Public License.
  */
 
-#define SETSERIAL_VERSION "2.14"
-#define SETSERIAL_DATE "9-Jun-98"
+#define SETSERIAL_VERSION "2.15"
+#define SETSERIAL_DATE "2-Nov-98"
